"""External-facing functions of Seedcase Sprout."""
# This exposes only the functions we want exposed when
# the package is imported via `from seedcase_sprout import *`.

from pprint import pprint
from textwrap import dedent

from .as_readme_text import as_readme_text
from .check_data import check_data
from .check_properties import (
    check_package_properties,
    check_properties,
    check_resource_properties,
)
from .create_properties_script import create_properties_script
from .create_resource_properties_script import create_resource_properties_script
from .examples import (
    ExamplePackage,
    example_data,
    example_data_all_types,
    example_package_properties,
    example_resource_properties,
    example_resource_properties_all_types,
)
from .extract_field_properties import extract_field_properties
from .join_resource_batches import join_resource_batches
from .paths import PackagePath
from .properties import (
    ConstraintsProperties,
    ContributorProperties,
    FieldProperties,
    LicenseProperties,
    PackageProperties,
    ReferenceProperties,
    ResourceProperties,
    SourceProperties,
    TableSchemaForeignKeyProperties,
    TableSchemaProperties,
)
from .read_properties import read_properties
from .read_resource_batches import read_resource_batches
from .read_resource_data import read_resource_data
from .write_file import write_file
from .write_properties import write_properties
from .write_resource_batch import write_resource_batch
from .write_resource_data import write_resource_data

__all__ = [
    # Properties -----
    "ConstraintsProperties",
    "ContributorProperties",
    "FieldProperties",
    "LicenseProperties",
    "PackageProperties",
    "ReferenceProperties",
    "ResourceProperties",
    "SourceProperties",
    "TableSchemaForeignKeyProperties",
    "TableSchemaProperties",
    "read_properties",
    "create_properties_script",
    # Example data and properties -----
    "example_package_properties",
    "example_package_properties_diabetes_study",
    "example_resource_properties",
    "example_data",
    "example_data_patients",
    "example_data_all_types",
    "example_resource_properties_all_types",
    "ExamplePackage",
    # Packages -----
    "write_properties",
    "as_readme_text",
    # Resources -----
    "extract_field_properties",
    "join_resource_batches",
    "read_resource_batches",
    "read_resource_data",
    "write_resource_batch",
    "create_resource_properties_script",
    "write_resource_data",
    # Path -----
    "PackagePath",
    # Helpers -----
    # "pretty_json",
    "write_file",
    "pprint",
    "dedent",
    # Checks -----
    "check_package_properties",
    "check_properties",
    "check_resource_properties",
    "check_data",
]



import re
from json import loads
from pathlib import Path
from typing import Any, Iterator

from jsonschema import Draft7Validator, FormatChecker, ValidationError

from seedcase_sprout.check_datapackage.check_error import CheckError
from seedcase_sprout.check_datapackage.constants import (
    COMPLEX_VALIDATORS,
    NAME_PATTERN,
    PACKAGE_RECOMMENDED_FIELDS,
    SEMVER_PATTERN,
)


def _read_json(path: Path) -> dict[str, Any]:
    """Reads the contents of a JSON file into an object.

    Args:
        path: The path to the file to load.

    Returns:
        The contents of the file as an object.

    Raises:
        JSONDecodeError: If the contents of the file cannot be de-serialised as JSON.
        TypeError: If the object in the file is not a dictionary.
    """
    loaded_object = loads(path.read_text())
    if not isinstance(loaded_object, dict):
        raise TypeError(
            f"Expected {path} to contain a JSON dictionary object "
            f"but found {type(loaded_object)}."
        )
    return loaded_object


def _add_package_recommendations(schema: dict[str, Any]) -> dict[str, Any]:
    """Add recommendations from the Data Package standard to the schema.

    Modifies the schema in place.

    Args:
        schema: The full Data Package schema.

    Returns:
        The updated Data Package schema.
    """
    schema["required"].extend(PACKAGE_RECOMMENDED_FIELDS.keys())
    schema["properties"]["name"]["pattern"] = NAME_PATTERN
    schema["properties"]["version"]["pattern"] = SEMVER_PATTERN
    schema["properties"]["contributors"]["items"]["required"] = ["title"]
    schema["properties"]["sources"]["items"]["required"] = ["title"]
    return schema


def _add_resource_recommendations(schema: dict[str, Any]) -> dict[str, Any]:
    """Add recommendations from the Data Resource standard to the schema.

    Modifies the schema in place.

    Args:
        schema: The full Data Package schema.

    Returns:
        The updated Data Package schema.
    """
    schema["properties"]["resources"]["items"]["properties"]["name"]["pattern"] = (
        NAME_PATTERN
    )
    return schema


def _check_object_against_json_schema(
    json_object: dict[str, Any], schema: dict[str, Any]
) -> list[CheckError]:
    """Checks that `json_object` matches the given JSON schema.

    Structural, type and format constraints are all checked. All schema violations are
    collected before errors are returned.

    Args:
        json_object: The JSON object to check.
        schema: The JSON schema to check against.

    Returns:
        A list of errors. An empty list, if no errors are found.

    Raises:
        jsonschema.exceptions.SchemaError: If the given schema is invalid.
    """
    Draft7Validator.check_schema(schema)
    validator = Draft7Validator(schema, format_checker=FormatChecker())
    return _validation_errors_to_check_errors(validator.iter_errors(json_object))


def _validation_errors_to_check_errors(
    validation_errors: Iterator[ValidationError],
) -> list[CheckError]:
    """Transforms `jsonschema.ValidationError`s to more compact `CheckError`s.

    The list of errors is:

      - flattened
      - filtered for summary-type errors
      - filtered for duplicates
      - sorted by error location

    Args:
        validation_errors: The `jsonschema.ValidationError`s to transform.

    Returns:
        A list of `CheckError`s.
    """
    check_errors = [
        CheckError(
            message=error.message,
            json_path=_get_full_json_path_from_error(error),
            validator=str(error.validator),
        )
        for error in _unwrap_errors(list(validation_errors))
        if str(error.validator) not in COMPLEX_VALIDATORS
    ]
    return sorted(set(check_errors))


def _unwrap_errors(errors: list[ValidationError]) -> list[ValidationError]:
    """Recursively extracts all errors into a flat list of errors.

    Args:
        errors: A nested list of errors.

    Returns:
        A flat list of errors.
    """
    unwrapped = []
    for error in errors:
        unwrapped.append(error)
        if error.context:
            unwrapped.extend(_unwrap_errors(error.context))
    return unwrapped


def _get_full_json_path_from_error(error: ValidationError) -> str:
    """Returns the full `json_path` to the error.

    For 'required' errors, the field name is extracted from the error message.

    Args:
        error: The error to get the full `json_path` for.

    Returns:
        The full `json_path` of the error.
    """
    if str(error.validator) == "required":
        match = re.search("'(.*)' is a required property", error.message)
        if match:
            return f"{error.json_path}.{match.group(1)}"
    return error.json_path



from typing import Any

from seedcase_sprout.check_datapackage.check_error import CheckError
from seedcase_sprout.check_datapackage.constants import DATA_PACKAGE_SCHEMA_PATH
from seedcase_sprout.check_datapackage.internals import (
    _add_resource_recommendations,
    _check_object_against_json_schema,
    _read_json,
)


def check_resource_properties(
    properties: dict[str, Any], check_recommendations: bool = True
) -> list[CheckError]:
    """Checks that the resource `properties` matches the Data Resource standard (v2.0).

    This function expects an individual set of resource properties as input. Structural,
    type and format constraints are all checked. All schema violations are collected
    before errors are returned.

    The schema loaded or constructed in this function overrides any values specified
    in the `$schema` attribute of `properties`, including the default value.

    Args:
        properties: The resource properties to check.
        check_recommendations: Whether `properties` should be checked against
            recommendations in the Data Resource standard in addition to requirements.
            Defaults to True.

    Returns:
        A list of errors. An empty list, if no errors are found.
    """
    schema = _read_json(DATA_PACKAGE_SCHEMA_PATH)

    # Recommendations from the Data Package standard
    if check_recommendations:
        _add_resource_recommendations(schema)

    # Consider only Data Resource schema
    schema = schema["properties"]["resources"]["items"]
    return _check_object_against_json_schema(properties, schema)



from pathlib import Path

from jinja2 import Environment, FileSystemLoader

from seedcase_sprout.constants import TEMPLATES_PATH
from seedcase_sprout.internals import _create_resource_properties_script_filename
from seedcase_sprout.paths import PackagePath
from seedcase_sprout.properties import FieldProperties
from seedcase_sprout.sprout_checks.is_resource_name_correct import (
    _is_resource_name_correct,
)
from seedcase_sprout.write_file import write_file


def create_resource_properties_script(
    resource_name: str | None = None,
    fields: list[FieldProperties] | None = None,
    path: Path | None = None,
) -> Path:
    """Create a script using the resource properties template.

    You can include the resource name and the fields' name and type information.
    If the script already exists, it will not be overwritten.

    Args:
        resource_name: The name of the new resource. Defaults to None.
        fields: The fields (columns) of the new resource. Defaults to None.
        path: The path to the package folder. Defaults to the current working directory.

    Returns:
        The path to the newly created script file.

    Raises:
        ValueError: If the resource name is not correct (e.g., contains spaces).

    Examples:
        ```{python}
        import seedcase_sprout as sp

        with sp.ExamplePackage():
            sp.create_resource_properties_script("my-resource")
        ```
    """
    if resource_name and not _is_resource_name_correct(resource_name):
        raise ValueError(
            f"The resource name '{resource_name}' is not correct. Resource names"
            " should only include lowercase alphanumeric characters and `.-_`."
        )
    resource_name = resource_name or ""
    script_path = PackagePath(path).resource_properties_script(resource_name)
    script_path.parent.mkdir(exist_ok=True)
    if script_path.exists():
        return script_path

    env = Environment(loader=FileSystemLoader(TEMPLATES_PATH), autoescape=True)
    env.filters["to_variable_name"] = _create_resource_properties_script_filename
    template = env.get_template("resource_properties.py.jinja2")
    text = template.render(resource_name=resource_name, fields=fields)

    return write_file(text, script_path)



# ruff: noqa
# mypy: ignore-errors
email  # unused variable (src/seedcase_sprout/properties.py:91)
given_name  # unused variable (src/seedcase_sprout/properties.py:92)
family_name  # unused variable (src/seedcase_sprout/properties.py:93)
organization  # unused variable (src/seedcase_sprout/properties.py:94)
roles  # unused variable (src/seedcase_sprout/properties.py:95)
email  # unused variable (src/seedcase_sprout/properties.py:137)
reference  # unused variable (src/seedcase_sprout/properties.py:173)
label  # unused variable (src/seedcase_sprout/properties.py:186)
required  # unused variable (src/seedcase_sprout/properties.py:242)
pattern  # unused variable (src/seedcase_sprout/properties.py:244)
enum  # unused variable (src/seedcase_sprout/properties.py:245)
min_length  # unused variable (src/seedcase_sprout/properties.py:246)
max_length  # unused variable (src/seedcase_sprout/properties.py:247)
minimum  # unused variable (src/seedcase_sprout/properties.py:248)
maximum  # unused variable (src/seedcase_sprout/properties.py:249)
exclusive_minimum  # unused variable (src/seedcase_sprout/properties.py:250)
exclusive_maximum  # unused variable (src/seedcase_sprout/properties.py:251)
json_schema  # unused variable (src/seedcase_sprout/properties.py:252)
example  # unused variable (src/seedcase_sprout/properties.py:287)
constraints  # unused variable (src/seedcase_sprout/properties.py:288)
categories  # unused variable (src/seedcase_sprout/properties.py:289)
categories_ordered  # unused variable (src/seedcase_sprout/properties.py:290)
fields_match  # unused variable (src/seedcase_sprout/properties.py:332)
unique_keys  # unused variable (src/seedcase_sprout/properties.py:334)
foreign_keys  # unused variable (src/seedcase_sprout/properties.py:335)
missing_values  # unused variable (src/seedcase_sprout/properties.py:336)
mediatype  # unused variable (src/seedcase_sprout/properties.py:391)
encoding  # unused variable (src/seedcase_sprout/properties.py:392)
bytes  # unused variable (src/seedcase_sprout/properties.py:393)
error_type  # unused variable (src/seedcase_sprout/examples.py:395)
error  # unused variable (src/seedcase_sprout/examples.py:396)
traceback  # unused variable (src/seedcase_sprout/examples.py:397)
none  # unused variable (tests/test_get_nested_attr.py:7)
Number_1  # unused variable (tests/test_get_nested_attr.py:8)



from pathlib import Path
from typing import cast

from seedcase_sprout.check_properties import check_properties
from seedcase_sprout.internals import _to_dedented, _write_json
from seedcase_sprout.internals.get import _get_nested_attr
from seedcase_sprout.paths import PackagePath
from seedcase_sprout.properties import (
    FieldProperties,
    PackageProperties,
)


def write_properties(properties: PackageProperties, path: Path | None = None) -> Path:
    """Write the `properties` to the `datapackage.json` file.

    If the `datapackage.json` file already exists, it will be overwritten. If not,
    a new file will be created.

    Args:
        properties: The properties to write. Use `create_properties_script()` to
            create a file with your properties object.
        path: A `Path` to the `datapackage.json` file.

    Returns:
        The path to the updated `datapackage.json` file.

    Raises:
        ExceptionGroup: If there is an error in the properties. A group of
            `CheckError`s, one error for each failed check.
    """
    path = path or PackagePath().properties()

    # Dedent descriptions
    properties.description = _to_dedented(properties.description)

    # TODO: Code to find all description fields and dedent to avoid nested for-loops?
    for resource in properties.resources or []:
        resource.description = _to_dedented(resource.description)

        for field in cast(
            list[FieldProperties],
            _get_nested_attr(resource, "schema.fields", default=[]),
        ):
            field.description = _to_dedented(field.description)

    check_properties(properties)

    return _write_json(properties.compact_dict, path)



from pytest import mark

from seedcase_sprout import PackagePath


def test_package_path_outputs_an_absolute_path(tmp_path):
    """Test that the path functions output an absolute path."""
    # This is a simple test to just check that a path is output.
    # Need to give the `tmp_path` to the class otherwise it uses the working directory
    # of the test script.
    path = PackagePath(tmp_path)
    assert path.root().is_absolute()
    assert path.properties().is_absolute()
    assert path.readme().is_absolute()
    assert path.resources().is_absolute()
    assert path.resource("test").is_absolute()
    assert path.resource_data("test").is_absolute()
    assert path.resource_batch("test").is_absolute()


def test_methods_return_correct_path(tmp_path):
    """Methods should return the expected path."""
    package_path = PackagePath(tmp_path)
    assert package_path.root() == tmp_path
    assert package_path.properties() == tmp_path / "datapackage.json"
    assert package_path.readme() == tmp_path / "README.md"
    assert package_path.resources() == tmp_path / "resources"
    assert package_path.resource("test") == tmp_path / "resources" / "test"
    assert (
        package_path.resource_data("test")
        == tmp_path / "resources" / "test" / "data.parquet"
    )

    assert (
        package_path.resource_batch("test") == tmp_path / "resources" / "test" / "batch"
    )


def test_resource_batch_files_returns_empty_list_when_no_batches(tmp_path):
    """resource_batch_files() should return an empty list when no batches are found
    for the resource."""
    assert PackagePath(tmp_path).resource_batch_files("test") == []


def test_resource_batch_files_returns_file_paths_when_batches(tmp_path):
    """resource_batch_files() should return the file paths to the batch files of the
    resource. Only Parquet files should be returned."""
    package_path = PackagePath(tmp_path)
    # Add batches for 2 resources
    for resource in ["test1", "test2"]:
        batch_folder = package_path.resource_batch(resource)
        batch_folder.mkdir(parents=True)
        (batch_folder / "sub-folder").mkdir()
        for file in ["file", "file.txt", "file.parquet"]:
            (batch_folder / file).touch()

    # Only the batch file for the given resource should be returned
    assert package_path.resource_batch_files("test2") == [batch_folder / "file.parquet"]


def test_path_defaults_to_cwd_at_call_time(tmp_cwd):
    """When no root path is provided, the root path should default to the cwd of the
    calling script."""
    package_path = PackagePath()
    assert package_path.root() == tmp_cwd


@mark.parametrize("resource_name", ["my_resource", "my.resource", "my-resource"])
def test_resource_properties_script_returns_correct_path_with_resource_name(
    tmp_path, resource_name
):
    """Should return the path to the resource properties script with the resource
    name in snake case."""
    package_path = PackagePath(tmp_path)
    assert (
        package_path.resource_properties_script(resource_name)
        == tmp_path / "scripts" / "resource_properties_my_resource.py"
    )


def test_resource_properties_script_returns_correct_path_without_resource_name(
    tmp_path,
):
    """Should return the path to the resource properties script without a
    resource name."""
    package_path = PackagePath(tmp_path)
    assert (
        package_path.resource_properties_script()
        == tmp_path / "scripts" / "resource_properties.py"
    )



from seedcase_sprout.check_datapackage.check_resource_properties import (
    check_resource_properties,
)

# Without recommendations


def test_passes_matching_properties():
    """Should pass properties matching the schema."""
    properties = {"name": "a name with spaces", "path": "data.csv"}

    assert check_resource_properties(properties, check_recommendations=False) == []


def test_fails_properties_with_missing_required_fields():
    """Should fail properties with missing required fields."""
    errors = check_resource_properties({}, check_recommendations=False)

    assert len(errors) == 3
    assert all(error.validator == "required" for error in errors)


def test_fails_properties_with_bad_type():
    """Should fail properties with a field of the wrong type."""
    properties = {"name": 123, "path": "data.csv"}

    errors = check_resource_properties(properties, check_recommendations=False)

    assert len(errors) == 1
    assert errors[0].validator == "type"
    assert errors[0].json_path == "$.name"


def test_fails_properties_with_bad_format():
    """Should fail properties with a field of the wrong format."""
    properties = {"name": "name", "path": "data.csv", "homepage": "not a URL"}

    errors = check_resource_properties(properties, check_recommendations=False)

    assert len(errors) == 1
    assert errors[0].validator == "format"
    assert errors[0].json_path == "$.homepage"


def test_fails_properties_with_pattern_mismatch():
    """Should fail properties with a field that does not match the pattern."""
    properties = {"name": "a name with spaces", "path": "/bad/path.csv"}

    errors = check_resource_properties(properties, check_recommendations=False)

    assert all(error.json_path == "$.path" for error in errors)
    assert any(error.validator == "pattern" for error in errors)


# With recommendations


def test_passes_matching_properties_with_recommendations():
    """Should pass properties matching recommendations."""
    properties = {"name": "a-name-with-no-spaces", "path": "data.csv"}

    assert check_resource_properties(properties, check_recommendations=True) == []


def test_fails_properties_with_pattern_mismatch_with_recommendations():
    """Should fail properties with field violating recommendations."""
    properties = {"name": "a name with spaces", "path": "data.csv"}

    errors = check_resource_properties(properties, check_recommendations=True)

    assert len(errors) == 1
    assert errors[0].validator == "pattern"
    assert errors[0].json_path == "$.name"



# ruff: noqa
# type: ignore
# NOTE: This content was autogenerated, no need to edit it.


from dataclasses import dataclass, field
from typing import Any, Literal


@dataclass
class ContributorProperties:
    """A contributor to this descriptor.

    Attributes:
        title (str | None): A human-readable title.
        path (str | None): A fully qualified URL, or a POSIX file path.
        email (str | None): An email address.
        given_name (str | None):
        family_name (str | None):
        organization (str | None): An organizational affiliation for this
            contributor.
        roles (list[str] | None):
    """

    title: str | None = None
    path: str | None = None
    email: str | None = None
    given_name: str | None = None
    family_name: str | None = None
    organization: str | None = None
    roles: list[str] | None = None


@dataclass
class LicenseProperties:
    """A license for this descriptor.

    Attributes:
        name (str | None): MUST be an Open Definition license identifier, see
            http://licenses.opendefinition.org/
        path (str | None): A fully qualified URL, or a POSIX file path.
        title (str | None): A human-readable title.
    """

    name: str | None = None
    path: str | None = None
    title: str | None = None


Path = str | None


@dataclass
class SourceProperties:
    """A source file.

    Attributes:
        title (str | None): A human-readable title.
        path (str | None): A fully qualified URL, or a POSIX file path.
        email (str | None): An email address.
        version (str | None):
    """

    title: str | None = None
    path: str | None = None
    email: str | None = None
    version: str | None = None


@dataclass
class TableDialectProperties:
    """The Table dialect descriptor.

    Attributes:
        header (bool | None): Specifies if the file includes a header row,
            always as the first row in the file.
        header_rows (list[int] | None):
        header_join (str | None):
        comment_rows (list[int] | None):
        comment_char (str | None): Specifies that any row beginning with this
            one-character string, without preceding whitespace, causes the entire
            line to be ignored.
        delimiter (str | None): A character sequence to use as the field
            separator.
        line_terminator (str | None): Specifies the character sequence that
            must be used to terminate rows.
        quote_char (str | None): Specifies a one-character string to use as
            the quoting character.
        double_quote (bool | None): Specifies the handling of quotes inside
            fields.
        escape_char (str | None): Specifies a one-character string to use as
            the escape character.
        null_sequence (str | None): Specifies the null sequence, for example,
            `\\N`.
        skip_initial_space (bool | None): Specifies the interpretation of
            whitespace immediately following a delimiter. If false, whitespace
            immediately after a delimiter should be treated as part of the
            subsequent field.
        property (str | None):
        item_type (Literal['array', 'object'] | None):
        item_keys (list[str] | None):
        sheet_number (int | None):
        sheet_name (str | None):
        table (str | None):
    """

    header: bool | None = True
    header_rows: list[int] | None = field(default_factory=lambda: [1])
    header_join: str | None = " "
    comment_rows: list[int] | None = field(default_factory=lambda: [1])
    comment_char: str | None = None
    delimiter: str | None = ","
    line_terminator: str | None = "\r\n"
    quote_char: str | None = '"'
    double_quote: bool | None = True
    escape_char: str | None = None
    null_sequence: str | None = None
    skip_initial_space: bool | None = False
    property: str | None = None
    item_type: Literal["array", "object"] | None = None
    item_keys: list[str] | None = None
    sheet_number: int | None = None
    sheet_name: str | None = None
    table: str | None = None


UniqueKey = list[str] | None


@dataclass
class ReferenceProperties:
    resource: str | None = None
    fields: list[str] | None = None


@dataclass
class TableSchemaForeignKeyProperties:
    """Table Schema Foreign Key

    Attributes:
        fields (list[str] | None):
        reference (ReferenceProperties | None):
    """

    fields: list[str] | None = None
    reference: ReferenceProperties | None = None


@dataclass
class MissingValueProperties:
    value: str | None = None
    label: str | None = None


@dataclass
class TableSchemaProperties:
    """A Table Schema for this resource, compliant with the [Table
    Schema](/tableschema/) specification.

    Attributes:
        fields_match (list | None):
        primary_key (list[str] | str | None): A primary key is a field name or
            an array of field names, whose values `MUST` uniquely identify each
            row in the table.
        unique_keys (list[UniqueKey] | None):
        foreign_keys (list[TableSchemaForeignKeyProperties] | None):
        missing_values (list[str] | list[MissingValueProperties] | None): Values that
            when encountered in the source, should be considered as `null`, 'not
            present', or 'blank' values.
    """

    fields_match: list | None = None
    primary_key: list[str] | str | None = None
    unique_keys: list[UniqueKey] | None = None
    foreign_keys: list[TableSchemaForeignKeyProperties] | None = None
    missing_values: list[str] | list[MissingValueProperties] | None = field(
        default_factory=lambda: [""]
    )


@dataclass
class ResourceProperties:
    """Data Resource.

    Attributes:
        name (str | None): An identifier string.
        path (str | list[Path] | None): A reference to the data for this
            resource, as either a path as a string, or an array of paths as
            strings. of valid URIs.
        data (Any | None): Inline data for this resource.
        type (Literal['table'] | None):
        title (str | None): A human-readable title.
        description (str | None): A text description. Markdown is encouraged.
        homepage (str | None): The home on the web that is related to this
            data package.
        sources (list[SourceProperties] | None): The raw sources for this resource.
        licenses (list[LicenseProperties] | None): The license(s) under which the
            resource is published.
        format (str | None): The file format of this resource.
        mediatype (str | None): The media type of this resource. Can be any
            valid media type listed with
            [IANA](https://www.iana.org/assignments/media-types/media-
            types.xhtml).
        encoding (str | None): The file encoding of this resource.
        bytes (int | None): The size of this resource in bytes.
        hash (str | None): The MD5 hash of this resource. Indicate other
            hashing algorithms with the {algorithm}:{hash} format.
        dialect (TableDialectProperties | None): The Table dialect descriptor.
        schema (TableSchemaProperties | None): A Table Schema for this resource,
            compliant with the [Table Schema](/tableschema/) specification.
    """

    name: str | None = None
    path: str | list[Path] | None = None
    data: Any | None = None
    type: Literal["table"] | None = None
    title: str | None = None
    description: str | None = None
    homepage: str | None = None
    sources: list[SourceProperties] | None = None
    licenses: list[LicenseProperties] | None = None
    format: str | None = None
    mediatype: str | None = None
    encoding: str | None = "utf-8"
    bytes: int | None = None
    hash: str | None = None
    dialect: TableDialectProperties | None = None
    schema: TableSchemaProperties | None = None


@dataclass
class PackageProperties:
    """Data Package

    Attributes:
        name (str | None): An identifier string.
        id (str | None): A property reserved for globally unique identifiers.
            Examples of identifiers that are unique include UUIDs and DOIs.
        title (str | None): A human-readable title.
        description (str | None): A text description. Markdown is encouraged.
        homepage (str | None): The home on the web that is related to this
            data package.
        version (str | None): A unique version number for this descriptor.
        created (str | None): The datetime on which this descriptor was
            created.
        contributors (list[ContributorProperties] | None): The contributors to this
            descriptor.
        keywords (list[str] | None): A list of keywords that describe this
            package.
        image (str | None): A image to represent this package.
        licenses (list[LicenseProperties] | None): The license(s) under which this
            package is published.
        resources (list[ResourceProperties] | None): An `array` of Data Resource
            objects, each compliant with the [Data Resource](/data-resource/)
            specification.
        sources (list[SourceProperties] | None): The raw sources for this resource.
    """

    name: str | None = None
    id: str | None = None
    title: str | None = None
    description: str | None = None
    homepage: str | None = None
    version: str | None = None
    created: str | None = None
    contributors: list[ContributorProperties] | None = None
    keywords: list[str] | None = None
    image: str | None = None
    licenses: list[LicenseProperties] | None = None
    resources: list[ResourceProperties] | None = None
    sources: list[SourceProperties] | None = None



from pathlib import Path


def write_file(string: str, path: Path) -> Path:
    """Write the `string` to a file at the specified `path`.

    The parent folder of the file in `path` must exist.
    If the file already exists, it will be overwritten.

    Args:
        string: The content you want written to the file.
        path: The full path to the file you want to create, including the file name and
            extension.

    Returns:
        The path to the file that was created.

    Raises:
        FileNotFoundError: If the parent folder of the file doesn't exist.
    """
    path.write_text(string)

    return path



from pathlib import Path

from jinja2 import Environment, FileSystemLoader

from seedcase_sprout.constants import TEMPLATES_PATH
from seedcase_sprout.paths import PackagePath
from seedcase_sprout.properties import PackageProperties
from seedcase_sprout.write_file import write_file


def create_properties_script(path: Path | None = None) -> Path:
    """Create a properties script with default values.

    If the script already exists, it will not be overwritten.

    Args:
        path: The path to the package folder. Defaults to the current working directory.

    Returns:
        The path to the newly created properties script.

    Examples:
        ```{python}
        import seedcase_sprout as sp

        with sp.ExamplePackage():
            sp.create_properties_script()
        ```
    """
    package_path = PackagePath(path)
    script_path = package_path.properties_script()
    script_path.parent.mkdir(exist_ok=True)
    # We don't want to overwrite an existing script.
    if script_path.exists():
        return script_path

    env = Environment(loader=FileSystemLoader(TEMPLATES_PATH), autoescape=True)
    template = env.get_template("package_properties.py.jinja2")
    text = template.render(
        properties=PackageProperties.from_default(name=package_path.root().name)
    )
    return write_file(text, script_path)



from typing import Any

from seedcase_sprout.check_datapackage import CheckError
from seedcase_sprout.sprout_checks.check_fields_not_blank import (
    check_fields_not_blank,
)
from seedcase_sprout.sprout_checks.check_list_item_field_not_blank import (
    check_list_item_field_not_blank,
)
from seedcase_sprout.sprout_checks.required_fields import (
    PACKAGE_SPROUT_REQUIRED_FIELDS,
)


def check_required_package_properties_not_blank(
    properties: dict[str, Any],
) -> list[CheckError]:
    """Checks that required package properties fields are not blank.

    Both Sprout-specific required fields and fields required by the Data Package
    standard are checked.

    Args:
        properties: The package properties.

    Returns:
        A list of errors. An empty list if no errors were found.
    """
    errors = check_fields_not_blank(properties, PACKAGE_SPROUT_REQUIRED_FIELDS)
    errors += check_list_item_field_not_blank(properties, "contributors", "title")
    errors += check_list_item_field_not_blank(properties, "sources", "title")
    errors += check_list_item_field_not_blank(properties, "licenses", "name")
    errors += check_list_item_field_not_blank(properties, "licenses", "path")
    return errors



from pytest import mark, raises

from seedcase_sprout.internals import _get_nested_attr


class D:
    none: str | None = None
    Number_1: int = 123


d = D()


class C:
    d: D = d


c = C()


class B:
    c: C = c


b = B()


class A:
    b: B = b


a = A()


@mark.parametrize(
    "nested_object,attributes,expected",
    [
        (a, "b.c.d", d),
        (a, "b.c", c),
        (a, "b", b),
        (b, "c.d", d),
        (c, "d", d),
        (d, "Number_1", 123),
        (d, "none", None),
        (a, "a", None),
        (a, "e", None),
        (a, "b.c.d.e", None),
        (a, "b.e.f.g", None),
        (a, "b.c.d.e.f.g.h", None),
    ],
)
def test_gets_nested_attribute(nested_object, attributes, expected):
    """Should resolve an attribute chain and return the correct value."""
    assert _get_nested_attr(nested_object, attributes) == expected


def test_returns_default_if_attribute_not_found():
    """Should return the default value when the attribute chain cannot be resolved."""
    assert _get_nested_attr(a, "e", default="default") == "default"


def test_returns_default_if_attribute_is_none():
    """Should return the default value when the attribute exists and is None."""
    assert _get_nested_attr(d, "none", default="default") == "default"


def test_ignores_default_if_attribute_found():
    """Should ignore the default value when the attribute chain can be resolved."""
    assert _get_nested_attr(d, "Number_1", default=456) == 123


@mark.parametrize(
    "attributes",
    [
        "",
        ".",
        "...",
        ".b",
        "...b",
        "b.",
        "b...",
        ".b.",
        ".b.c.d",
        "b..c",
        ".b.c.d.",
        "b,c,d",
        "$",
        "1",
    ],
)
def test_throws_error_if_attribute_not_identifier(attributes):
    """Should throw `ValueError` if the input doesn't contain valid identifiers."""
    with raises(ValueError):
        _get_nested_attr(a, attributes)



import re
from dataclasses import dataclass

from seedcase_sprout.check_datapackage.check_error import CheckError


@dataclass
class CheckErrorMatcher:
    r"""A class that helps to filter `CheckError`s based on their attributes.

    Examples:
            ```{python}
            error = CheckError(
                message="123 is not of type 'string'",
                json_path="$.resources[0].name",
                validator="type",
            )
            matcher = CheckErrorMatcher(
                message="of type 'string'",
                json_path=r"resources\[.\]",
                validator="type",
            )
            matcher.matches(error)

            ```
    """

    # The substring to find in the error message.
    message: str | None = None
    # The subpath to find in the error's `json_path`. Expressed as a regular expression.
    json_path: str | None = None
    # The validator to match.
    validator: str | None = None

    def matches(self, error: CheckError) -> bool:
        """Determines if this matcher matches the given `CheckError`.

        Args:
            error: The `CheckError` to match.

        Returns:
            If there was a match.
        """
        return (
            self.message_matches(error)
            and self.json_path_matches(error)
            and self.validator_matches(error)
        )

    def message_matches(self, error: CheckError) -> bool:
        """Determines if this matcher matches the message of the given `CheckError`.

        Args:
            error: The `CheckError` to match.

        Returns:
            If there was a match.
        """
        return self.message is None or self.message in error.message

    def json_path_matches(self, error: CheckError) -> bool:
        """Determines if this matcher matches the `json_path` of the given `CheckError`.

        Matching regular expressions is supported.

        Args:
            error: The `CheckError` to match.

        Returns:
            If there was a match.
        """
        if self.json_path is None:
            return True
        return re.search(self.json_path, error.json_path) is not None

    def validator_matches(self, error: CheckError) -> bool:
        """Determines if this matcher matches the validator of the given `CheckError`.

        Args:
            error: The `CheckError` to match.

        Returns:
            If there was a match.
        """
        return self.validator is None or self.validator == error.validator



from typing import cast

import polars as pl

from seedcase_sprout.properties import FieldType

"""
Mapping between Polars types and Data Package types.

Each Polars type can be described by any of the specified Data Package types.
When extracting Data Package types from a data frame, each Polars type is mapped to
the first Data Package type in the corresponding entry.
"""
_POLARS_TO_DATAPACKAGE: dict[type[pl.DataType], list[FieldType]] = {
    pl.Int8: ["integer", "year", "any"],
    pl.Int16: ["integer", "year", "any"],
    pl.Int32: ["integer", "year", "any"],
    pl.Int64: ["integer", "year", "any"],
    pl.Int128: ["integer", "year", "any"],
    pl.UInt8: ["integer", "year", "any"],
    pl.UInt16: ["integer", "year", "any"],
    pl.UInt32: ["integer", "year", "any"],
    pl.UInt64: ["integer", "year", "any"],
    pl.Float32: ["number", "any"],
    pl.Float64: ["number", "any"],
    pl.Decimal: ["number", "any"],
    pl.Boolean: ["boolean", "any"],
    pl.String: ["string", "array", "object", "geojson", "duration", "any"],
    pl.Binary: ["string", "any"],
    pl.Categorical: ["string", "any"],
    pl.Enum: ["string", "any"],
    pl.Date: ["date", "yearmonth", "any"],
    pl.Datetime: ["datetime", "any"],
    pl.Time: ["time", "any"],
    pl.Duration: ["integer", "any"],
    pl.Array: ["array", "any"],
    pl.List: ["array", "any"],
    pl.Struct: ["object", "geojson", "any"],
    pl.Object: ["object", "geojson", "any"],
    pl.Unknown: ["any"],
    pl.Null: ["any"],
}


def _get_allowed_datapackage_types(polars_type: pl.DataType) -> list[FieldType]:
    """Return the Data Package types that can describe the given Polars type.

    Args:
        polars_type: The Polars type to get the Data Package types for.

    Returns:
        The allowed Data Package types.
    """
    base_type = cast(type[pl.DataType], polars_type.base_type())
    allowed_types = _POLARS_TO_DATAPACKAGE.get(base_type, ["any"])

    if (
        isinstance(polars_type, pl.Array)
        and polars_type.size == 2
        and polars_type.inner.is_numeric()
    ):
        return allowed_types + ["geopoint"]

    return allowed_types


def _polars_and_datapackage_types_match(
    polars_type: pl.DataType, datapackage_type: FieldType | None
) -> bool:
    """Decide if the given Polars and Data Package types match.

    Args:
        polars_type: The Polars type to check.
        datapackage_type: The Data Package type to check.

    Returns:
        Whether the given types match.
    """
    return (datapackage_type or "any") in _get_allowed_datapackage_types(polars_type)


def _get_allowed_polars_types(
    datapackage_type: FieldType | None,
) -> list[type[pl.DataType]]:
    """Return the Polars types that can represent the given Data Package type.

    Args:
        datapackage_type: The Data Package type.

    Returns:
        The allowed Polars types.
    """
    datapackage_type = datapackage_type or "any"
    if datapackage_type == "geopoint":
        # Strictly speaking, this should be an Array of a numeric type with size 2, but
        # it wouldn't be practical to list all combinations here. This information is
        # added directly to the error message instead.
        return [pl.Array]

    # TODO: Revise once `_map_keep()` has been implemented
    return [
        polars_type
        for polars_type, datapackage_types in _POLARS_TO_DATAPACKAGE.items()
        if datapackage_type in datapackage_types
    ]


def _polars_to_datapackage(polars_type: pl.DataType) -> FieldType:
    """Return the Data Package type that is the best match for the given Polars type.

    Args:
        polars_type: The Polars type to match.

    Returns:
        The best-match Data Package type.
    """
    return _get_allowed_datapackage_types(polars_type)[0]



from pathlib import Path

import polars as pl

from seedcase_sprout.check_data import check_data
from seedcase_sprout.paths import PackagePath
from seedcase_sprout.properties import ResourceProperties


def write_resource_data(
    data: pl.DataFrame,
    resource_properties: ResourceProperties,
    package_path: Path | None = None,
) -> Path:
    """Check and write the resource data into a file.

    This function takes the `data` obtained after using
    `join_resource_batches()`, checks it against the `resource_properties`, and
    then writes the data to the resources `data.parquet` file .  The Parquet
    file is saved based on the path found in `ResourceProperties.path` and is
    always overwritten.  Before writing, this function does a check against the
    `resource_properties` to ensure that the data is correctly structured and
    tidy.

    Args:
        data: A DataFrame object with the resources data from the files in its
            `batch/` folder.
        resource_properties: The `ResourceProperties` object that contains the
            properties of the resource you want to create the Parquet file for.
        package_path: The path to the data package root folder (where `datapackage.json`
            is located). Defaults to the current working directory.

    Returns:
        The path of the created Parquet file.

    Examples:
        ```{python}
        import seedcase_sprout as sp

        with sp.ExamplePackage():
            resource_properties = sp.example_resource_properties()
            # Add and join batch files
            sp.write_resource_batch(sp.example_data(), resource_properties)
            batches = sp.read_resource_batches(resource_properties)
            data = sp.join_resource_batches(batches, resource_properties)
            # Write resource data file
            sp.write_resource_data(data, resource_properties)
    """
    check_data(data, resource_properties)
    data_path = PackagePath(package_path).resource_data(str(resource_properties.name))

    data.write_parquet(data_path)
    return data_path



from typing import Any

from seedcase_sprout.check_datapackage.check_error import CheckError
from seedcase_sprout.check_datapackage.constants import (
    DATA_PACKAGE_SCHEMA_PATH,
)
from seedcase_sprout.check_datapackage.internals import (
    _add_package_recommendations,
    _check_object_against_json_schema,
    _read_json,
)


def check_package_properties(
    properties: dict[str, Any], check_recommendations: bool = True
) -> list[CheckError]:
    """Checks that `properties` matches the Data Package standard (v2.0).

    Only package properties are checked. Schema constraints for resource properties are
    removed, so the internal structure of resource properties is not checked.
    Structural, type and format constraints are all checked. All schema violations are
    collected before errors are returned.

    The schema loaded or constructed in this function overrides any values specified
    in the `$schema` attribute of `properties`, including the default value.

    Args:
        properties: The package properties to check.
        check_recommendations: Whether `properties` should be checked against
            recommendations in the Data Package standard in addition to requirements.
            Defaults to True.

    Returns:
        A list of errors. An empty list, if no errors are found.
    """
    schema = _read_json(DATA_PACKAGE_SCHEMA_PATH)

    # Recommendations from the Data Package standard
    if check_recommendations:
        _add_package_recommendations(schema)

    # Remove schema constraints for resource properties
    schema["required"].remove("resources")
    del schema["properties"]["resources"]["minItems"]
    del schema["properties"]["resources"]["items"]

    return _check_object_against_json_schema(properties, schema)



import re

from datamodel_code_generator.format import CustomCodeFormatter


class CodeFormatter(CustomCodeFormatter):
    """A class for postprocessing the code generated by datamodel-codegen."""

    def apply(self, code: str) -> str:
        """Applies postprocessing logic to the generated class file.

        Args:
            code: the generated code.

        Returns:
            The code after postprocessing.
        """
        # append 'Properties' to class names in class definitions and type hints
        class_pattern = r"class (\w+)"
        class_names = re.findall(class_pattern, code)

        for class_name in class_names:
            class_name_pattern = rf"\b{class_name}\b"
            code = re.sub(class_name_pattern, f"{class_name}Properties", code)

        # miscellaneous edits
        code = code.replace("DataPackageProperties", "PackageProperties")
        code = code.replace("DataResourceProperties", "ResourceProperties")
        code = code.replace("schema_", "schema")
        code = code.replace("from __future__ import annotations\n", "")
        code = code.replace(r"\N", r"\\N")

        return code



import os
import tempfile
from contextlib import AbstractContextManager
from pathlib import Path
from types import TracebackType

import polars as pl

from seedcase_sprout.as_readme_text import as_readme_text
from seedcase_sprout.paths import PackagePath
from seedcase_sprout.properties import (
    ContributorProperties,
    FieldProperties,
    LicenseProperties,
    PackageProperties,
    ResourceProperties,
    TableSchemaProperties,
)
from seedcase_sprout.write_file import write_file
from seedcase_sprout.write_properties import write_properties


def example_package_properties() -> PackageProperties:
    """Generate an example package properties object.

    Returns:
        An example `PackageProperties` object.

    Examples:
        ```{python}
        import seedcase_sprout as sp
        sp.example_package_properties()
        ```
    """
    properties = PackageProperties.from_default(
        name="example-package",
        title="Example fake data package",
        description="Data from a fake data package on something.",
        contributors=[
            ContributorProperties(
                title="Jamie Jones",
                email="jamie_jones@example.com",
                roles=["creator"],
            )
        ],
        licenses=[
            LicenseProperties(
                name="ODC-BY-1.0",
                path="https://opendatacommons.org/licenses/by",
                title="Open Data Commons Attribution License 1.0",
            )
        ],
    )

    return properties


def example_resource_properties() -> ResourceProperties:
    """Generate an example resource properties object.

    Returns:
        Outputs a correctly formatted example `ResourceProperties` object.

    Examples:
        ```{python}
        import seedcase_sprout as sp
        sp.example_resource_properties()
        ```
    """
    return ResourceProperties(
        name="example-resource",
        title="Example fake data resource",
        description="Data from a fake resource on something.",
        schema=TableSchemaProperties(
            fields=[
                FieldProperties(
                    name="id",
                    title="ID",
                    description="Unique identifier for the record.",
                    type="integer",
                ),
                FieldProperties(
                    name="name",
                    title="Name",
                    description="Name of the record.",
                    type="string",
                ),
                FieldProperties(
                    name="value",
                    title="Value",
                    description="Value of the record.",
                    type="number",
                ),
            ],
            primary_key=["id"],
        ),
    )


def example_data() -> pl.DataFrame:
    """Generate an example Polars data frame.

    Returns:
        A Polars data frame matching the resource properties generated by
            `example_resource_properties()`.

    Examples:
        ```{python}
        import seedcase_sprout as sp
        sp.example_data()
        ```
    """
    return pl.DataFrame(
        {
            "id": [34, 99, 100],
            "name": ["Helly R", "Mark S", "Ms Casey"],
            "value": [123.123, 9988, -76.0009],
        }
    )


def example_resource_properties_all_types() -> ResourceProperties:
    """Generate an example resource properties object with all Data Package data types.

    Returns:
        Outputs a correctly formatted example `ResourceProperties` object.

    Examples:
        ```{python}
        import seedcase_sprout as sp
        sp.example_resource_properties_all_types()
        ```
    """
    return ResourceProperties(
        name="data",
        title="data",
        description="My data...",
        schema=TableSchemaProperties(
            fields=[
                FieldProperties(name="my_boolean", type="boolean"),
                FieldProperties(name="my_integer", type="integer"),
                FieldProperties(name="my_number", type="number"),
                FieldProperties(name="my_year", type="year"),
                FieldProperties(name="my_yearmonth", type="yearmonth"),
                FieldProperties(name="my_datetime_tz", type="datetime"),
                FieldProperties(name="my_datetime_no_tz", type="datetime"),
                FieldProperties(name="my_date", type="date"),
                FieldProperties(name="my_time", type="time"),
                FieldProperties(name="my_geopoint", type="geopoint"),
                FieldProperties(name="my_array", type="array"),
                FieldProperties(name="my_object", type="object"),
                FieldProperties(name="my_string", type="string"),
                FieldProperties(name="my_any", type="any"),
                FieldProperties(name="my_none", type=None),
                FieldProperties(name="my_duration", type="duration"),
                FieldProperties(name="my_geojson", type="geojson"),
            ],
        ),
    )


def example_data_all_types() -> pl.DataFrame:
    """Generate an example Polars data frame with all Data Package data types.

    Returns:
        A Polars data frame matching the resource properties generated by
            `example_resource_properties_all_types()`.

    Examples:
        ```{python}
        import seedcase_sprout as sp
        sp.example_data_all_types()
        ```
    """
    return pl.DataFrame(
        {
            "my_boolean": [True, True, False],
            "my_integer": [12223, -123, 0],
            "my_number": [123.123, float("inf"), float("2E-33")],
            "my_year": [2012, -1000, 0],
            "my_yearmonth": pl.Series(
                ["2014-12-01", "0001-01-01", "0000-01-01"]
            ).str.to_date(),
            "my_datetime_tz": pl.Series(
                [
                    "2002-10-10T12:00:00+01:00",
                    "2002-05-12T12:40:00.34+01:00",
                    "2019-01-20T09:30:00+01:00",
                ]
            ).str.to_datetime(),
            "my_datetime_no_tz": pl.Series(
                [
                    "2002-10-10T12:44:10",
                    "2002-10-10T12:00:00.34",
                    "0000-01-01T00:00:00",
                ]
            ).str.to_datetime(),
            "my_date": pl.Series(
                ["2002-10-10", "0001-01-01", "0000-01-01"]
            ).str.to_date(),
            "my_time": pl.Series(
                ["15:00:59", "00:00:00.3", "12:00:00.345345"]
            ).str.to_time(),
            "my_geopoint": pl.Series([[-90, -180], [5, 45], [5.9999, 45.0000]]).cast(
                pl.Array(pl.Float64, 2)
            ),
            "my_array": [
                "[]",
                '[{"prop1": true}]',
                '[{"prop1": "value"}, {"prop2": 123}]',
            ],
            "my_object": [
                "{}",
                '{"outer": "value", "inner": {"prop1": 123, "prop2": [1, 2, null]}}',
                '{"outer": "value", "inner": {"prop1": 123, "prop2": true}}',
            ],
            "my_string": ["some text", "æøåäöü£$%^&*()\\''", "μῆνιν ἄειδε θεὰ"],
            "my_any": pl.Series([[1]] * 3, dtype=pl.List),
            "my_none": pl.Series([[1]] * 3, dtype=pl.List),
            "my_duration": ["P1Y2M3DT10H30M45.343S", "P1Y2M3DT10H30M45", "P33Y6M4D"],
            "my_geojson": ["{}"] * 3,
        }
    )


def example_resource_properties_all_polars_types() -> ResourceProperties:
    """Generate an example resource properties object with all Polars data types.

    Returns:
        Outputs a correctly formatted example `ResourceProperties` object.
        sp.example_resource_properties_all_polars_types()
        ```
    """
    return ResourceProperties(
        name="example-resource",
        title="data",
        path=str(Path("resources", "example-resource", "data.parquet")),
        description="My data...",
        schema=TableSchemaProperties(
            fields=[
                FieldProperties(name="my_decimal", type="number"),
                FieldProperties(name="my_float32", type="number"),
                FieldProperties(name="my_float64", type="number"),
                FieldProperties(name="my_int8", type="integer"),
                FieldProperties(name="my_int16", type="integer"),
                FieldProperties(name="my_int32", type="integer"),
                FieldProperties(name="my_int64", type="integer"),
                FieldProperties(name="my_int128", type="integer"),
                FieldProperties(name="my_uint8", type="integer"),
                FieldProperties(name="my_uint16", type="integer"),
                FieldProperties(name="my_uint32", type="integer"),
                FieldProperties(name="my_uint64", type="integer"),
                FieldProperties(name="my_date", type="date"),
                FieldProperties(name="my_datetime", type="datetime"),
                FieldProperties(name="my_duration", type="integer"),
                FieldProperties(name="my_time", type="time"),
                FieldProperties(name="my_array", type="array"),
                # We can't map to DataPackage `list` bc it's missing in the TableSchema
                # profile we use to generate the datapackage schema that we check
                # properties against. We've made an issue about this in the DataPackage
                # repo.
                FieldProperties(name="my_list", type="array"),
                FieldProperties(name="my_struct", type="object"),
                FieldProperties(name="my_string", type="string"),
                FieldProperties(name="my_categorical", type="string"),
                FieldProperties(name="my_enum", type="string"),
                FieldProperties(name="my_binary", type="string"),
                FieldProperties(name="my_boolean", type="boolean"),
                FieldProperties(name="my_null", type="any"),
                FieldProperties(name="my_object", type="object"),
            ],
        ),
    )


def example_data_all_polars_types() -> pl.DataFrame:
    """Generate an example Polars data frame with all Polars data types.

    Returns:
        A Polars data frame matching the resource properties generated by
            `example_resource_properties_all_polars_types()`.

    Examples:
        ```{python}
        import seedcase_sprout as sp
        sp.example_data_all_polars_types()
        ```
    """
    return pl.DataFrame(
        {
            "my_decimal": pl.Series([1.1, 2.2], dtype=pl.Decimal(10, 2)),
            "my_float32": pl.Series([1.1, 2.2], dtype=pl.Float32),
            "my_float64": pl.Series([1.1, 2.2], dtype=pl.Float64),
            "my_int8": pl.Series([1, 2], dtype=pl.Int8),
            "my_int16": pl.Series([1, 2], dtype=pl.Int16),
            "my_int32": pl.Series([1, 2], dtype=pl.Int32),
            "my_int64": pl.Series([1, 2], dtype=pl.Int64),
            "my_int128": pl.Series([1, 2], dtype=pl.Int128),
            "my_uint8": pl.Series([1, 2], dtype=pl.UInt8),
            "my_uint16": pl.Series([1, 2], dtype=pl.UInt16),
            "my_uint32": pl.Series([1, 2], dtype=pl.UInt32),
            "my_uint64": pl.Series([1, 2], dtype=pl.UInt64),
            "my_date": pl.Series(["2023-01-01", "2023-01-02"]).str.to_date(),
            "my_datetime": pl.Series(
                ["2023-01-01 00:00:00", "2023-01-02 00:00:00"]
            ).str.to_datetime(),
            "my_duration": pl.Series([1, 2]).cast(pl.Duration("ms")),
            "my_time": pl.Series(["00:00:00", "01:00:00"]).str.to_time(),
            "my_array": pl.Series([[1, 2], [3, 4]]).cast(pl.Array(pl.Int32, shape=2)),
            "my_list": pl.Series([[1, 2], [3, 4]]).cast(pl.List(pl.Int32)),
            "my_struct": pl.Series([{"a": 1}, {"b": 2}]),
            "my_string": pl.Series(["a", "b"]),
            "my_categorical": pl.Series(["a", "b"]).cast(pl.Categorical),
            "my_enum": pl.Series(["a", "b"]).cast(pl.Enum(categories=["a", "b"])),
            # pl.Utf8 is an alias for pl.String, so it's not included here
            "my_binary": pl.Series([b"a", b"b"]),
            "my_boolean": pl.Series([True, False]),
            "my_null": pl.Series([None, None]),
            "my_object": pl.Series(["a", {"a": 1}], dtype=pl.Object),
            # TODO: Couldn't find a way to create pl.Unknown
        }
    )


class ExamplePackage(AbstractContextManager[PackagePath]):
    """Create a temporary data package with optional resources for demoing or testing.

    Examples:
        ```{python}
        import seedcase_sprout as sp
        with sp.ExamplePackage() as package_path:
            properties = sp.read_properties(package_path.properties())

        with sp.ExamplePackage(with_resources=False):
            properties = sp.read_properties()

        with sp.ExamplePackage(package_name="my-package"):
            properties = sp.read_properties()
        ```
    """

    def __init__(
        self,
        package_name: str = str(example_package_properties().name),
        with_resources: bool = True,
    ):
        """Initialise the `ExamplePackage` context manager.

        Args:
            package_name: The name of the package and its root folder.
            with_resources: Whether resources should be added when creating the package.
                Defaults to True.
        """
        self.package_name = package_name
        self.with_resources = with_resources
        self.calling_dir = Path.cwd()
        self.temp_dir = tempfile.TemporaryDirectory()

    def __enter__(self) -> PackagePath:
        """Create the temporary package structure and switch to its directory.

        Returns:
            A `PackagePath` object pointing to the root of the temporary package.
        """
        package_dir = Path(self.temp_dir.name) / self.package_name
        package_dir.mkdir(exist_ok=True)
        package_path = PackagePath(package_dir)

        # Create package properties
        package_properties = example_package_properties()
        package_properties.name = self.package_name

        # Create resource properties
        if self.with_resources:
            resource_properties = example_resource_properties()
            package_properties.resources = [resource_properties]

            # Create resource folders
            package_path.resource(str(resource_properties.name)).mkdir(
                exist_ok=True, parents=True
            )

            # Include some data in the resource
            example_data().write_parquet(
                package_path.resource_data(str(resource_properties.name)),
            )

        # Save properties
        write_properties(properties=package_properties, path=package_path.properties())

        # Write README
        write_file(as_readme_text(package_properties), package_path.readme())

        os.chdir(package_path.root())

        return package_path

    def __exit__(
        self,
        error_type: type[BaseException] | None,
        error: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Restore the original working directory and clean up the temporary package."""
        os.chdir(self.calling_dir)
        self.temp_dir.cleanup()



from pathlib import Path

import polars as pl

from seedcase_sprout.paths import PackagePath


def read_resource_data(
    resource_name: str,
    path: Path | None = None,
) -> pl.DataFrame:
    """Read the resource's `data.parquet` file.

    Use this function to read in the data file for a specific resource in a data
    package as a Polars DataFrame.

    Args:
        resource_name: The name of the resource. This should match the name
            found in the `datapackage.json` file under the `resource.name`
            key for the resource.
        path: The path to the data package folder, where the `datapackage.json`
            file is located. If not provided, it defaults to the current
            working directory. Use `PackagePath().properties()` to help get the
            correct path.

    Returns:
        A Polars DataFrame containing the data from the resource's data file.

    Raises:
        FileNotFoundError: If the resource's data file does not exist in the
            specified path.

    Examples:
        ```{python}
        import seedcase_sprout as sp
        with sp.ExamplePackage() as package_path:
            data = sp.read_resource_data("example-resource")
            print(data)
        ```
    """
    data_path = PackagePath(path).resource_data(resource_name)
    if not data_path.is_file():
        raise FileNotFoundError(
            f"Resource '{resource_name}' not found in the package, so we couldn't",
            "read the data.",
        )

    return pl.read_parquet(data_path)



from pathlib import Path
from typing import cast

from pytest import ExceptionInfo, fixture, mark, raises

from seedcase_sprout.check_datapackage import CheckError
from seedcase_sprout.check_properties import (
    check_package_properties,
    check_properties,
    check_resource_properties,
)
from seedcase_sprout.examples import example_resource_properties
from seedcase_sprout.properties import (
    ContributorProperties,
    LicenseProperties,
    PackageProperties,
    ResourceProperties,
    SourceProperties,
)
from seedcase_sprout.sprout_checks.get_blank_value_for_type import (
    get_blank_value_for_type,
)
from seedcase_sprout.sprout_checks.required_fields import (
    PACKAGE_SPROUT_REQUIRED_FIELDS,
    RESOURCE_SPROUT_REQUIRED_FIELDS,
)


@fixture
def properties():
    return PackageProperties(
        name="package-1",
        id="abc1",
        title="Package 1",
        description="A package.",
        version="1.0.0",
        created="2024-05-14T05:00:01+00:00",
        licenses=[LicenseProperties(name="a-license")],
        contributors=[ContributorProperties(title="a contributor")],
        sources=[SourceProperties(title="a source")],
        resources=[
            ResourceProperties(
                name="resource-1",
                title="Resource 1",
                description="A resource.",
            ),
            ResourceProperties(
                name="resource-2",
                title="Resource 2",
                description="A second resource.",
            ),
        ],
    )


def test_passes_correct_properties(properties):
    """Should pass if all required properties are present and correct."""
    assert check_properties(properties) == properties
    assert check_package_properties(properties) == properties
    assert check_resource_properties(properties.resources[0]) == properties.resources[0]
    assert check_resource_properties(properties.resources[1]) == properties.resources[1]

    # Even when resources isn't there.
    delattr(properties, "resources")
    assert check_properties(properties) == properties


def test_error_incorrect_argument():
    """Should be an error if it isn't a `*Properties` object."""
    with raises(TypeError):
        check_properties(ResourceProperties())

    with raises(TypeError):
        check_resource_properties(PackageProperties())

    with raises(TypeError):
        check_properties("")

    with raises(TypeError):
        check_properties([1, 2, 3])

    with raises(TypeError):
        check_package_properties([1, 2, 3])

    with raises(TypeError):
        check_package_properties("")

    with raises(TypeError):
        check_resource_properties([1, 2, 3])

    with raises(TypeError):
        check_resource_properties("")


@mark.parametrize("field", PACKAGE_SPROUT_REQUIRED_FIELDS.keys())
def test_error_missing_required_package_properties(properties, field):
    """Should be an error if a required package properties is missing."""
    delattr(properties, field)

    # All properties checks
    with raises(ExceptionGroup) as error_info:
        check_properties(properties)

    errors = _as_check_errors(error_info)
    assert len(errors) == 1
    assert errors[0].json_path == f"$.{field}"
    assert errors[0].validator == "required"

    # Package only checks
    with raises(ExceptionGroup) as error_info:
        check_package_properties(properties)

    errors = _as_check_errors(error_info)
    assert len(errors) == 1
    assert errors[0].json_path == f"$.{field}"
    assert errors[0].validator == "required"


@mark.parametrize(
    "item,value,validator",
    [
        ("name", "a name with spaces", "pattern"),
        ("title", 123, "type"),
        ("homepage", "not a URL", "format"),
        ("resources", 123, "type"),
    ],
)
def test_error_incorrect_property_values(properties, item, value, validator):
    """Should be an error when the property value is incorrect."""
    setattr(properties, item, value)

    with raises(ExceptionGroup) as error_info:
        check_package_properties(properties)

    errors = _as_check_errors(error_info)
    assert len(errors) == 1
    assert errors[0].json_path == f"$.{item}"
    assert errors[0].validator == f"{validator}"

    with raises(ExceptionGroup) as error_info:
        check_properties(properties)

    errors = _as_check_errors(error_info)
    assert len(errors) == 1
    assert errors[0].json_path == f"$.{item}"
    assert errors[0].validator == f"{validator}"


@mark.parametrize("name,type", PACKAGE_SPROUT_REQUIRED_FIELDS.items())
def test_error_blank_package_properties(properties, name, type):
    """Should be an error when a required package field is blank."""
    setattr(properties, name, get_blank_value_for_type(type))

    with raises(ExceptionGroup) as error_info:
        check_properties(properties)

    blank_errors = [
        error for error in _as_check_errors(error_info) if error.validator == "blank"
    ]

    assert len(blank_errors) == 1
    assert blank_errors[0].json_path == f"$.{name}"

    with raises(ExceptionGroup) as error_info:
        check_package_properties(properties)

    blank_errors = [
        error for error in _as_check_errors(error_info) if error.validator == "blank"
    ]

    assert len(blank_errors) == 1
    assert blank_errors[0].json_path == f"$.{name}"


def test_error_missing_required_nested_properties(properties):
    """Should have errors when the nested required properties are missing."""
    setattr(properties, "licenses", [{}])
    setattr(properties, "contributors", [{}])
    setattr(properties, "sources", [{}])

    with raises(ExceptionGroup) as error_info:
        check_package_properties(properties)

    required_errors = [
        error for error in _as_check_errors(error_info) if error.validator == "required"
    ]
    assert [error.json_path for error in required_errors] == [
        "$.contributors[0].title",
        "$.licenses[0].name",
        "$.licenses[0].path",
        "$.sources[0].title",
    ]

    with raises(ExceptionGroup) as error_info:
        check_properties(properties)

    required_errors = [
        error for error in _as_check_errors(error_info) if error.validator == "required"
    ]
    assert [error.json_path for error in required_errors] == [
        "$.contributors[0].title",
        "$.licenses[0].name",
        "$.licenses[0].path",
        "$.sources[0].title",
    ]


# Resource properties specific --------------------------------------------


@mark.parametrize(
    "field",
    RESOURCE_SPROUT_REQUIRED_FIELDS.keys(),
)
def test_error_missing_required_resource_properties(properties, field):
    """Should be an error if a required resource properties is missing."""
    delattr(properties.resources[0], field)

    with raises(ExceptionGroup) as error_info:
        check_resource_properties(properties.resources[0])

    errors = _as_check_errors(error_info)
    assert len(errors) == 1
    assert errors[0].json_path == f"$.{field}"
    assert errors[0].validator == "required"

    with raises(ExceptionGroup) as error_info:
        check_properties(properties)

    errors = _as_check_errors(error_info)
    assert len(errors) == 1
    assert errors[0].json_path == f"$.resources[0].{field}"
    assert errors[0].validator == "required"


@mark.parametrize("name,type", RESOURCE_SPROUT_REQUIRED_FIELDS.items())
def test_error_blank_resource_properties(properties, name, type):
    """Should be an error when one required resource field is blank."""
    setattr(properties.resources[0], name, get_blank_value_for_type(type))

    with raises(ExceptionGroup) as error_info:
        check_properties(properties)

    blank_errors = [
        error for error in _as_check_errors(error_info) if error.validator == "blank"
    ]

    assert len(blank_errors) == 1
    assert blank_errors[0].json_path == f"$.resources[0].{name}"

    with raises(ExceptionGroup) as error_info:
        check_resource_properties(properties.resources[0])

    blank_errors = [
        error for error in _as_check_errors(error_info) if error.validator == "blank"
    ]

    assert len(blank_errors) == 1
    assert blank_errors[0].json_path == f"$.{name}"


def test_errors_flagged_for_fields_with_multipart_name():
    """Errors should be flagged when the name of the field has more than one word."""
    properties = example_resource_properties()
    assert properties.schema
    properties.schema.primary_key = []

    with raises(ExceptionGroup) as error_info:
        check_resource_properties(properties)

    assert all(
        error.json_path == "$.schema.primaryKey"
        for error in _as_check_errors(error_info)
    )


def test_error_incorrect_resource_property_values(properties):
    """Should be an error when the property value is incorrect."""
    properties.resources[0].title = 123

    with raises(ExceptionGroup) as error_info:
        check_resource_properties(properties.resources[0])

    errors = _as_check_errors(error_info)
    assert len(errors) == 1
    assert errors[0].json_path == "$.title"
    assert errors[0].validator == "type"

    with raises(ExceptionGroup) as error_info:
        check_properties(properties)

    errors = _as_check_errors(error_info)
    assert len(errors) == 1
    assert errors[0].json_path == "$.resources[0].title"
    assert errors[0].validator == "type"


@mark.parametrize(
    "path", ["", [], 123, str(Path("resources", "1")), "/bad/path/data.csv"]
)
def test_error_no_resource_name_in_path(properties, path):
    """Should be an error when the resource name isn't in the `path` or is empty."""
    properties.resources[0].path = path

    with raises(ExceptionGroup) as error_info:
        check_resource_properties(properties.resources[0])

    errors = _as_check_errors(error_info)
    assert len(errors) >= 1
    assert all(error.json_path.endswith("path") for error in errors)

    with raises(ExceptionGroup) as error_info:
        check_properties(properties)

    errors = _as_check_errors(error_info)
    assert len(errors) >= 1
    assert all(error.json_path.endswith("path") for error in errors)


def _as_check_errors(
    error_info: ExceptionInfo[ExceptionGroup[Exception]],
) -> list[CheckError]:
    errors = error_info.value.exceptions
    assert all(isinstance(error, CheckError) for error in errors)
    return cast(list[CheckError], errors)



import polars as pl

from seedcase_sprout.internals import _map, _map2
from seedcase_sprout.map_data_types import (
    _polars_to_datapackage,
)
from seedcase_sprout.properties import FieldProperties


def extract_field_properties(data: pl.DataFrame) -> list[FieldProperties]:
    """Extract field properties from a Polars DataFrame.

    Data types are extracted from the DataFrame's schema and mapped from Polars to
    Data Package data types. For the best results, ensure that the DataFrame's schema is
    correct---this may require some manually fixing. Since there are differences
    between Polars and Data Package data types, the mapping may not always be
    perfect, and some data types may be extracted to simpler types. For
    example, a year column may be extracted as an integer instead of Data
    Package's `year` type. Be sure to review and modify the extracted
    properties to suit your needs.

    Args:
        data: A Polars DataFrame containing the data to extract properties from.

    Returns:
        A list of `FieldProperties` objects, each representing a field/column in the
        DataFrame.

    Raises:
        ValueError: If the data is empty.

    Examples:
        ```{python}
        import seedcase_sprout as sp

        sp.extract_field_properties(
            data=sp.example_data(),
        )
        ```
    """
    if data.is_empty():
        raise ValueError("Data is empty. Cannot extract field properties.")

    # TODO: add format="binary" to Binary field type?
    field_names = data.columns
    field_types = _map(data.dtypes, _polars_to_datapackage)

    field_properties = _map2(
        field_names,
        field_types,
        lambda name, type: FieldProperties(name=name, type=type),
    )

    return field_properties



from pathlib import Path


def _check_is_dir(path: Path) -> Path:
    """Checks whether the path exists and is a directory.

    Args:
        path: The path to check.

    Returns:
        The path if it is a directory.

    Raises:
        NotADirectoryError: If path either doesn't exist or isn't a directory.
    """
    if not path.is_dir():
        raise NotADirectoryError(f"{path} either doesn't exist or isn't a directory.")

    return path


def _check_is_file(path: Path) -> Path:
    """Checks whether the file given by the path exists and is a file.

    Args:
        path: The path to check.

    Returns:
        A path to the file if the path refers to a file.

    Raises:
        FileNotFound: If the file in the path doesn't exist or isn't a file.
    """
    if not path.is_file():
        raise FileNotFoundError(f"{path} is not a file.")

    return path



from typing import Any

from seedcase_sprout.check_datapackage import CheckError
from seedcase_sprout.sprout_checks.get_json_path_to_resource_field import (
    get_json_path_to_resource_field,
)


def check_no_inline_data(
    properties: dict[str, Any], index: int | None = None
) -> list[CheckError]:
    """Checks that the `data` field of a set of resource properties is not set.

    Args:
        properties: The resource properties.
        index: The index of the resource properties. Defaults to None.

    Returns:
        A list of errors. The empty list if no error was found.
    """
    if properties.get("data") is None:
        return []

    return [
        CheckError(
            message=(
                "Sprout doesn't use the 'data' field, instead it expects data "
                "in separate files that are given in the 'path' field."
            ),
            json_path=get_json_path_to_resource_field("data", index),
            validator="inline-data",
        )
    ]



from pytest import mark

from seedcase_sprout.check_datapackage.check_package_properties import (
    check_package_properties,
)

# Without recommendations


def test_passes_matching_properties():
    """Should pass properties matching the schema."""
    properties = {
        "name": "a name with spaces",
        "title": "A Title",
        "created": "2024-05-14T05:00:01+00:00",
        "version": "a version",
        "contributors": [{"email": "jane@doe.com"}],
        "sources": [{"email": "jane@doe.com"}],
    }

    assert check_package_properties(properties, check_recommendations=False) == []


@mark.parametrize("resources", [[], [{}], [{"name": "name", "path": "data.csv"}]])
def test_passes_matching_properties_without_checking_resources(resources):
    """Should pass matching package properties without checking individual resource
    properties."""
    properties = {"resources": resources}

    assert check_package_properties(properties, check_recommendations=False) == []


def test_fails_properties_with_resources_of_wrong_type():
    """Should fail properties if they have a `resources` field with a value of the wrong
    type."""
    properties = {"resources": 123}
    errors = check_package_properties(properties, check_recommendations=False)

    assert len(errors) == 1
    assert errors[0].validator == "type"
    assert errors[0].json_path == "$.resources"


def test_fails_properties_with_missing_required_fields():
    """Should fail properties with missing required fields."""
    properties = {
        "name": "a name",
        "licenses": [{"title": "my license"}],
    }

    errors = check_package_properties(properties, check_recommendations=False)

    assert len(errors) == 2
    assert all(error.validator == "required" for error in errors)
    assert {error.json_path for error in errors} == {
        "$.licenses[0].name",
        "$.licenses[0].path",
    }


def test_fails_properties_with_bad_type():
    """Should fail properties with a field of the wrong type."""
    properties = {"name": 123}
    errors = check_package_properties(properties, check_recommendations=False)

    assert len(errors) == 1
    assert errors[0].validator == "type"
    assert errors[0].json_path == "$.name"


def test_fails_properties_with_bad_format():
    """Should fail properties with a field of the wrong format."""
    properties = {"name": "a name", "homepage": "not a URL"}

    errors = check_package_properties(properties, check_recommendations=False)

    assert len(errors) == 1
    assert errors[0].validator == "format"
    assert errors[0].json_path == "$.homepage"


def test_fails_properties_with_pattern_mismatch():
    """Should fail properties with a field that does not match the pattern."""
    properties = {"name": "a name", "contributors": [{"path": "/a/bad/path"}]}

    errors = check_package_properties(properties, check_recommendations=False)

    assert len(errors) == 1
    assert errors[0].validator == "pattern"
    assert errors[0].json_path == "$.contributors[0].path"


# With recommendations


def test_passes_matching_properties_with_recommendations():
    """Should pass properties matching recommendations."""
    properties = {
        "name": "a-name-with-no-spaces",
        "title": "A Title",
        "id": "123",
        "created": "2024-05-14T05:00:01+00:00",
        "version": "3.2.1",
        "contributors": [{"title": "a contributor"}],
        "sources": [{"title": "a source"}],
        "licenses": [{"name": "a-license"}],
    }

    assert check_package_properties(properties, check_recommendations=True) == []


def test_fails_properties_with_missing_required_fields_with_recommendations():
    """Should fail properties with missing required fields."""
    errors = check_package_properties({}, check_recommendations=True)

    assert len(errors) == 3
    assert all(error.validator == "required" for error in errors)


def test_fails_properties_violating_recommendations():
    """Should fail properties that do not meet the recommendations."""
    properties = {
        "name": "a name with spaces",
        "id": "123",
        "version": "not semver",
        "contributors": [{"email": "jane@doe.com"}],
        "sources": [{"email": "jane@doe.com"}],
        "licenses": [{"name": "a-license"}],
    }

    errors = check_package_properties(properties, check_recommendations=True)

    assert len(errors) == 4
    assert {error.json_path for error in errors} == {
        "$.name",
        "$.version",
        "$.contributors[0].title",
        "$.sources[0].title",
    }



from pytest import mark

from seedcase_sprout.sprout_checks.check_resource_path_string import (
    check_resource_path_string,
)
from seedcase_sprout.sprout_checks.get_json_path_to_resource_field import (
    get_json_path_to_resource_field,
)


@mark.parametrize("index", [None, 2])
def test_passes_if_data_path_string(index):
    """Should pass if the path is of type string."""
    properties = {"path": "a string"}

    assert check_resource_path_string(properties, index) == []


@mark.parametrize("index", [None, 2])
def test_passes_if_data_path_not_present(index):
    """Should pass if the path is not set."""
    assert check_resource_path_string({}, index) == []


@mark.parametrize("index", [None, 2])
def test_error_found_if_path_not_string(index):
    """Should find an error if the path is not of type string."""
    properties = {"path": 123}

    errors = check_resource_path_string(properties, index)

    assert len(errors) == 1
    assert "string" in errors[0].message
    assert errors[0].json_path == get_json_path_to_resource_field("path", index)
    assert errors[0].validator == "type"



import re
from datetime import datetime
from pathlib import Path

import polars as pl

from seedcase_sprout.check_data import check_data
from seedcase_sprout.check_properties import (
    check_resource_properties,
)
from seedcase_sprout.constants import (
    BATCH_TIMESTAMP_COLUMN_NAME,
    BATCH_TIMESTAMP_FORMAT,
    BATCH_TIMESTAMP_PATTERN,
)
from seedcase_sprout.internals import _check_is_file, _map, _map2
from seedcase_sprout.paths import PackagePath
from seedcase_sprout.properties import ResourceProperties


def read_resource_batches(
    resource_properties: ResourceProperties, paths: list[Path] | None = None
) -> list[pl.DataFrame]:
    """Read all the batch resource file(s) into a list of (Polars) DataFrames.

    Use this function to read the Parquet file(s) specified in `paths` into
    a list of Polars DataFrames, and perform checks on each of the DataFrames
    against the `resource_properties`. The `resource_properties` object is used
    to check the data and ensure it is correct. While Sprout generally assumes
    that the files stored in the `resources/<name>/batch/` folder are already
    correctly structured and tidy, this function still runs checks to ensure the
    data are correct by comparing to the properties.

    Args:
        resource_properties: The `ResourceProperties` object that contains the
            properties of the resource you want to check the data against.
        paths: A list of paths for all the Parquet files in the resource's `batch/`
            folder. Use `PackagePath().resource_batch_files()` to help provide the
            correct paths to the batch files. Defaults to the batch files of the given
            resource.

    Returns:
        A list of DataFrame objects from all the batch files.

    Raises:
        ValueError: If the batch file name is not in the expected pattern.
        ValueError: If the timestamp column name matches an existing column in the
            DataFrame.

    Examples:
        ``` {python}
        import seedcase_sprout as sp

        with sp.ExamplePackage():
            resource_properties = sp.example_resource_properties()
            sp.write_resource_batch(sp.example_data(), resource_properties)

            sp.read_resource_batches(resource_properties)
        ```
    """
    check_resource_properties(resource_properties)
    if paths is None:
        paths = PackagePath().resource_batch_files(str(resource_properties.name))

    _map(paths, _check_is_file)
    return _map2(paths, [resource_properties], _read_parquet_batch_file)


def _read_parquet_batch_file(
    path: Path, resource_properties: ResourceProperties
) -> pl.DataFrame:
    """Reads a Parquet batch file and adds the timestamp as a column.

    This function reads a Parquet batch file into a Polars DataFrame and adds
    a timestamp column to the DataFrame, extracted from the file name.

    Args:
        path: Path to the Parquet batch file.
        resource_properties: The resource properties to check the data against.

    Returns:
        The Parquet file as a DataFrame with a timestamp column added.
    """
    if path.suffix != ".parquet":
        raise ValueError(
            "Failed to read batch file. Expected a file with a "
            f"`.parquet` extension but found {path}."
        )
    data = pl.read_parquet(path)
    check_data(data, resource_properties)

    timestamp = _extract_timestamp_from_batch_file_path(path)
    _check_batch_file_timestamp(timestamp)
    data = _add_timestamp_as_column(data, timestamp)
    return data


def _extract_timestamp_from_batch_file_path(path: Path) -> str:
    """Extracts the timestamp from the file name.

    Since the batch file name has been created by `_create_batch_file_name()`,
    it should contain a timestamp in the format defined by BATCH_TIMESTAMP_PATTERN.

    If multiple timestamps are found in the file name, the first one is used.
    """
    timestamp_list: list[str] = re.findall(BATCH_TIMESTAMP_PATTERN, path.stem)

    if not timestamp_list:
        raise ValueError(
            f"Batch file name '{path.stem}' does not contain a timestamp in the "
            f"expected format '{BATCH_TIMESTAMP_PATTERN}'."
        )

    return timestamp_list[0]


def _check_batch_file_timestamp(timestamp: str) -> str:
    """Checks the timestamp format and that it is a correct calendar date."""
    try:
        datetime.strptime(timestamp, BATCH_TIMESTAMP_FORMAT)
        return timestamp
    except ValueError as error:
        raise ValueError(
            f"Timestamp '{timestamp}' in the batch file name is not in the "
            f"expected format '{BATCH_TIMESTAMP_FORMAT}' or is not a correct calendar "
            "date (e.g., 30 February)."
        ) from error


def _add_timestamp_as_column(data: pl.DataFrame, timestamp: str) -> pl.DataFrame:
    """Adds the timestamp as a column to the data.

    Args:
        data: Data to add timestamp column to.
        timestamp: Timestamp to add as values in the timestamp column.

    Returns:
        Data with added timestamp column.

    Raises:
        ValueError: If a column with the name BATCH_TIMESTAMP_COLUMN_NAME already exists
        in the data.
    """
    # TODO: We could move this to be a check of the resource properties in
    # `sprout_checks/`
    if BATCH_TIMESTAMP_COLUMN_NAME in data.columns:
        raise ValueError(
            "One or multiple of the provided resource batch files contain a "
            f"column named '{BATCH_TIMESTAMP_COLUMN_NAME}'. This column is used "
            "internally in Sprout to remove duplicate rows across batches. Please "
            "rename it in the batch files and resource properties to read the resource "
            "batches."
        )
    return data.with_columns(pl.lit(timestamp).alias(BATCH_TIMESTAMP_COLUMN_NAME))



from json import JSONDecodeError
from pathlib import Path

from pytest import raises

from seedcase_sprout import (
    example_package_properties,
    read_properties,
    write_properties,
)
from seedcase_sprout.internals import _write_json
from seedcase_sprout.paths import PackagePath


def test_reads_in_as_package_properties(tmp_path):
    """Should read in the properties from the `datapackage.json` file."""

    expected_properties = example_package_properties()
    properties_path = tmp_path / "datapackage.json"
    properties_path = write_properties(expected_properties, properties_path)
    actual_properties = read_properties(properties_path)

    assert expected_properties == actual_properties


def test_reads_when_resource_not_exists(tmp_path):
    """Should not give an error if there are no resources on the package."""

    expected_properties = example_package_properties()
    expected_properties.resources = None
    properties_path = tmp_path / "datapackage.json"
    properties_path = write_properties(expected_properties, properties_path)
    actual_properties = read_properties(properties_path)

    assert expected_properties == actual_properties


def test_throws_error_if_path_points_to_dir(tmp_path):
    """Should throw FileNotFoundError if the path points to a folder."""
    with raises(FileNotFoundError):
        read_properties(tmp_path)


def test_throws_error_if_path_points_to_nonexistent_file(tmp_path):
    """Should throw FileNotFoundError if the path points to a nonexistent file."""
    with raises(FileNotFoundError):
        read_properties(tmp_path / "datapackage.json")


def test_throws_error_if_properties_file_cannot_be_read(tmp_path):
    """Should throw JSONDecodeError if the properties file cannot be read as JSON."""
    file_path = Path(tmp_path / "datapackage.json")
    file_path.write_text(",,, this is not, JSON")

    with raises(JSONDecodeError):
        read_properties(file_path)


def test_error_incorrect_properties_in_file(tmp_path):
    """Can't read in properties if the properties file is incorrect."""
    properties = example_package_properties()
    properties.name = "incorrect name"
    _write_json(properties.compact_dict, tmp_path / "datapackage.json")

    with raises(ExceptionGroup):
        read_properties(tmp_path / "datapackage.json")


def test_reads_properties_from_cwd_if_no_path_provided(tmp_cwd):
    """If no path is provided, should read datapackage.json from the cwd."""
    properties = example_package_properties()
    write_properties(properties, PackagePath(tmp_cwd).properties())

    assert read_properties() == properties


def test_fails_correctly_if_no_path_provided_and_no_properties_in_cwd(tmp_cwd):
    """Should throw the expected error if no path is provided and there is no
    datapackage.json in the cwd."""
    with raises(FileNotFoundError):
        read_properties()
